from django.core.management.commands.showmigrations import (
    Command as ShowMigrationsCommand,
)
from django.db.migrations.loader import MigrationLoader
from django.db.migrations.recorder import MigrationRecorder

from django_migrations_tui.tui.app import Format, MigrationsApp
from django_migrations_tui.tui.utils import MigrationsList


class Command(ShowMigrationsCommand):
    help = "Show all migrations and their status in a TUI."

    def handle(self, *args, **options):
        self.verbosity = options["verbosity"]
        self.options = options

        if options["format"] == "plan":
            format = Format.PLAN
        else:
            format = Format.LIST

        app = MigrationsApp(format=format)
        app.run()

    def show_list(self, connection, app_names=None):
        """
        Show a list of all migrations on the system, or only those of
        some named apps.
        """
        # Load migrations from disk/DB
        loader = MigrationLoader(connection, ignore_no_migrations=True)
        recorder = MigrationRecorder(connection)
        recorded_migrations = recorder.applied_migrations()
        graph = loader.graph
        # If we were passed a list of apps, validate it
        if app_names:
            self._validate_app_names(loader, app_names)
        # Otherwise, show all apps in alphabetic order
        else:
            app_names = sorted(loader.migrated_apps)

        migrations = list()
        for app_name in app_names:
            app_migration = MigrationsList(app_name, 0, list())
            shown = set()
            for node in graph.leaf_nodes(app_name):
                for plan_node in graph.forwards_plan(node):
                    if plan_node not in shown and plan_node[0] == app_name:
                        # Give it a nice title if it's a squashed one
                        title = plan_node[1]
                        if graph.nodes[plan_node].replaces:
                            title += " (%s squashed migrations)" % len(
                                graph.nodes[plan_node].replaces
                            )
                        applied_migration = loader.applied_migrations.get(plan_node)
                        # Mark it as applied/unapplied
                        if applied_migration:
                            if plan_node in recorded_migrations:
                                output = " [X] %s" % title
                                app_migration.applied_count += 1
                            else:
                                title += " Run 'manage.py migrate' to finish recording."
                                output = " [-] %s" % title
                            if self.verbosity >= 2 and hasattr(
                                applied_migration, "applied"
                            ):
                                output += (
                                    " (applied at %s)"
                                    % applied_migration.applied.strftime(
                                        "%Y-%m-%d %H:%M:%S"
                                    )
                                )
                            app_migration.migrations.append(output)
                        else:
                            app_migration.migrations.append(" [ ] %s" % title)
                        shown.add(plan_node)
            # If we didn't print anything, then a small message
            if not shown:
                app_migration.migrations.append(" (no migrations)")

            migrations.append(app_migration)
        return migrations

    def show_plan(self, connection, app_names=None):
        """
        Show all known migrations (or only those of the specified app_names)
        in the order they will be applied.
        """
        # Load migrations from disk/DB
        loader = MigrationLoader(connection)
        graph = loader.graph
        if app_names:
            self._validate_app_names(loader, app_names)
            targets = [key for key in graph.leaf_nodes() if key[0] in app_names]
        else:
            targets = graph.leaf_nodes()
        plan = []
        seen = set()

        # Generate the plan
        for target in targets:
            for migration in graph.forwards_plan(target):
                if migration not in seen:
                    node = graph.node_map[migration]
                    plan.append(node)
                    seen.add(migration)

        # Output
        def print_deps(node):
            out = []
            for parent in sorted(node.parents):
                out.append("%s.%s" % parent.key)
            if out:
                return " ... (%s)" % ", ".join(out)
            return ""

        migrations_plan = []
        for node in plan:
            deps = ""
            if self.verbosity >= 2:
                deps = print_deps(node)
            if node.key in loader.applied_migrations:
                migrations_plan.append(
                    "[X]  %s.%s%s" % (node.key[0], node.key[1], deps)
                )
            else:
                migrations_plan.append(
                    "[ ]  %s.%s%s" % (node.key[0], node.key[1], deps)
                )
        if not plan:
            migrations_plan.append("(no migrations)")

        return migrations_plan



import asyncio

from asgiref.sync import sync_to_async
from textual import work
from textual.message import Message
from textual.widgets import Tree
from textual.worker import Worker, WorkerState

from .utils import Format, get_migrations_list, get_migrations_plan

# TODO: Add a way to search for a migration and filter


class MigrationsTree(Tree):
    """A widget to display migrations."""

    class Status(Message):
        """A message to update the status of a migration."""

        def __init__(self, message: str):
            self.message = message
            super().__init__()

    def __init__(self, format: Format):
        self.format = format
        super().__init__("")

    def on_mount(self):
        self.styles.border = ("round", "#0178D4")

        if self.format == Format.PLAN:
            return self.update_migrations_plan()
        else:
            return self.update_migrations_list()

    async def toggle_format(self) -> None:
        if self.format == Format.LIST:
            self.format = Format.PLAN
            await self.update_migrations_plan()
        else:
            self.format = Format.LIST
            await self.update_migrations_list()

    async def update_migrations_list(self) -> None:
        migrations = await self.get_migrations_list()
        total_migrations = sum(
            [
                len(app.migrations)
                for app in migrations
                if app.migrations != [" (no migrations)"]
            ]
        )
        total_applied = sum([app.applied_count for app in migrations])

        self.reset("migrations (%s/%s)" % (total_applied, total_migrations))
        self.root.expand()
        for app in migrations:
            current_app = self.root.add(str(app))
            for migration_name in app.migrations:
                current_app.add_leaf(migration_name)

    async def reload_migrations_list(self) -> None:
        migrations = await self.get_migrations_list()
        total_migrations = sum(
            [
                len(app.migrations)
                for app in migrations
                if app.migrations != [" (no migrations)"]
            ]
        )
        total_applied = sum([app.applied_count for app in migrations])
        self.root.set_label("migrations (%s/%s)" % (total_applied, total_migrations))

        for child, app in zip(self.root.children, migrations):
            child.set_label(str(app))
            for child_child, migration_name in zip(child.children, app.migrations):
                child_child.set_label(migration_name)

    async def update_migrations_plan(self) -> None:
        migrations = await self.get_migrations_plan()
        total_migrations = len(migrations)
        total_applied = sum(
            [1 for migration in migrations if migration.startswith("[X]")]
        )

        self.reset("plan (%s/%s)" % (total_applied, total_migrations))
        self.root.expand()
        for item in migrations:
            self.root.add_leaf(item)

    @sync_to_async
    def get_migrations_list(self):
        return get_migrations_list()

    @sync_to_async
    def get_migrations_plan(self):
        return get_migrations_plan()

    def action_migrate(self):
        if self.format == Format.LIST:
            return self.migrate_list()
        else:
            return self.migrate_plan()

    def migrate_list(self):
        selected_item = self.cursor_node
        if selected_item.is_root:
            command = ["python", "manage.py", "migrate"]
        elif selected_item.allow_expand:
            app_name = str(selected_item.label).split(" (")[0]
            command = ["python", "manage.py", "migrate", app_name]
        elif (
            not selected_item.allow_expand
            and str(selected_item.label) == " (no migrations)"
        ):
            self.post_error_message("No migrations to apply.")
            return None
        else:
            app_name = str(selected_item.parent.label).split(" (")[0]
            migration_name = str(selected_item.label).split("] ")[1]
            command = ["python", "manage.py", "migrate", app_name, migration_name]

        return command

    def migrate_plan(self):
        selected_item = self.cursor_node
        if selected_item.is_root:
            command = ["python", "manage.py", "migrate"]
        else:
            # discard the [X] from the label
            migration = str(selected_item.label)[5:]
            command = ["python", "manage.py", "migrate", *migration.split(".")]

        return command

    @work(exclusive=True)
    async def run_command(self, command):
        self.post_success_message(f"Running {' '.join(command)}")

        proc = await asyncio.create_subprocess_exec(
            *command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        while proc.returncode is None:
            line = await proc.stdout.readline()
            if not line:
                break
            self.post_message(self.Status(line.decode("utf-8").rstrip()))
        _, stderr = await proc.communicate()
        if stderr:
            self.post_message(self.Status(stderr.decode("utf-8")))

    async def on_worker_state_changed(self, event: Worker.StateChanged) -> None:
        """Called when the worker state changes."""
        if event.state in [
            WorkerState.CANCELLED,
            WorkerState.ERROR,
            WorkerState.SUCCESS,
        ]:
            if self.format == Format.LIST:
                await self.reload_migrations_list()
            else:
                await self.update_migrations_plan()

    def fake_migration(self):
        if self.format == Format.LIST:
            return self.fake_migration_list()
        else:
            return self.fake_migration_plan()

    def fake_migration_list(self):
        selected_item = self.cursor_node
        if selected_item.is_root:
            command = ["python", "manage.py", "migrate", "--fake"]
        elif selected_item.allow_expand:
            app_name = str(selected_item.label).split(" (")[0]
            command = ["python", "manage.py", "migrate", "--fake", app_name]
        elif (
            not selected_item.allow_expand
            and str(selected_item.label) == " (no migrations)"
        ):
            self.post_error_message("No migrations to fake.")
            return None
        else:
            app_name = str(selected_item.parent.label).split(" (")[0]
            migration_name = str(selected_item.label).split("] ")[1]
            command = [
                "python",
                "manage.py",
                "migrate",
                "--fake",
                app_name,
                migration_name,
            ]
        return command

    def fake_migration_plan(self):
        selected_item = self.cursor_node
        if selected_item.is_root:
            command = ["python", "manage.py", "migrate", "--fake"]
        else:
            # discard the [X] from the label
            migration = str(selected_item.label)[5:]
            command = [
                "python",
                "manage.py",
                "migrate",
                "--fake",
                *migration.split("."),
            ]

        return command

    def revert_migrations(self):
        """Revert all migrations for an app."""
        if self.format == Format.LIST:
            selected_item = self.cursor_node
            if selected_item.allow_expand and not selected_item.is_root:
                app_name = str(selected_item.label).split(" (")[0]
                command = ["python", "manage.py", "migrate", app_name, "zero"]
                return command
            else:
                self.post_warning_message("Select an app to revert.")
        else:
            self.post_error_message("Revert not supported in plan format.")

    def post_warning_message(self, message: str) -> None:
        """Post a warning message to the log."""
        self.post_message(self.Status(f"[bold #ffa62b]{message}"))

    def post_error_message(self, message: str) -> None:
        """Post an error message to the log."""
        self.post_message(self.Status(f"[bold #ff0000]{message}"))

    def post_success_message(self, message: str) -> None:
        """Post a success message to the log."""
        self.post_message(self.Status(f"[bold cyan]{message}"))



import enum
import sys
from dataclasses import dataclass
from typing import List

from django.core.management import ManagementUtility, handle_default_options
from django.db import connections


class Format(enum.Enum):
    LIST = "list"
    PLAN = "plan"


@dataclass
class MigrationsList:
    app_name: str
    applied_count: int
    migrations: List[str]

    def __str__(self):
        if self.migrations == [" (no migrations)"]:
            return f"{self.app_name} (0/0)"
        return f"{self.app_name} ({self.applied_count}/{len(self.migrations)})"


def get_migrations(format: Format = Format.LIST, argv=None):
    utility = ManagementUtility(["migrationstui"])
    command = utility.fetch_command("migrationstui")

    if argv is None:
        argv = sys.argv
    if len(argv) < 2:
        argv = ["manage.py", "migrationstui"]

    command._called_from_command_line = True
    parser = command.create_parser(argv[0], argv[1])

    options = parser.parse_args(argv[2:])
    cmd_options = vars(options)
    # Move positional args out of options to mimic legacy optparse
    handle_default_options(options)
    command.verbosity = cmd_options["verbosity"]

    db = cmd_options["database"]
    connection = connections[db]

    if format == Format.LIST:
        migrations = command.show_list(connection, cmd_options["app_label"])
    else:
        migrations = command.show_plan(connection, cmd_options["app_label"])
    return migrations


def get_migrations_list(argv=None):
    return get_migrations(format=Format.LIST, argv=argv)


def get_migrations_plan(argv=None):
    return get_migrations(format=Format.PLAN, argv=argv)



from textual.app import App, ComposeResult
from textual.containers import Vertical
from textual.widgets import Footer, Header

from .confirmation import ConfirmationScreen
from .logs import Log
from .tree import MigrationsTree
from .utils import Format


class MigrationsApp(App):
    """A Textual app to manage django migrations."""

    CSS_PATH = "static/app.tcss"
    TITLE = "Django Migrations TUI"

    BINDINGS = [
        ("v", "toggle_format", "View"),
        ("l", "toggle_logs", "Logs"),
        ("m", "migrate", "Migrate"),
        ("f", "fake_migration", "Fake"),
        ("r", "revert_migrations", "Revert"),
    ]

    def __init__(self, *args, format: Format, **kwargs):
        self.format = format
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Header()
            yield MigrationsTree(self.format)
            yield Log(markup=True, highlight=True, wrap=True)
            yield Footer()

    def on_migrations_tree_status(self, message: MigrationsTree.Status) -> None:
        """Called when the status of a migration changes."""
        rich_log = self.query_one(Log)
        rich_log.write(message.message)
        rich_log.display = True

    def action_toggle_logs(self) -> None:
        """An action to toggle logs."""
        rich_log = self.query_one(Log)
        rich_log.display = not rich_log.display

    async def action_toggle_format(self) -> None:
        """An action to toggle the format of the migrations tree."""
        tree = self.query_one(MigrationsTree)
        await tree.toggle_format()

    def action_migrate(self) -> None:
        tree = self.query_one(MigrationsTree)
        command = tree.action_migrate()

        def check_confirmation(apply) -> None:
            if apply:
                tree.run_command(command)

        if command:
            self.push_screen(ConfirmationScreen(command), check_confirmation)

    def action_fake_migration(self) -> None:
        """An action to fake a migration."""
        tree = self.query_one(MigrationsTree)
        command = tree.fake_migration()

        def check_confirmation(apply) -> None:
            if apply:
                tree.run_command(command)

        if command:
            self.push_screen(ConfirmationScreen(command), check_confirmation)

    def action_revert_migrations(self) -> None:
        """An action to revert a migration."""
        tree = self.query_one(MigrationsTree)
        command = tree.revert_migrations()

        def check_confirmation(apply) -> None:
            if apply:
                tree.run_command(command)

        if command:
            self.push_screen(ConfirmationScreen(command), check_confirmation)



from typing import List
from textual.app import ComposeResult
from textual.containers import Grid
from textual.screen import ModalScreen
from textual.widgets import Button, Label


class ConfirmationScreen(ModalScreen[bool]):
    """Screen with a confirmation dialog"""

    def __init__(self, command: List[str]):
        super().__init__()
        self.command = f"[bold cyan]{' '.join(command)}"

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(self.command, id="question"),
            Button("Migrate", variant="primary", id="yes"),
            Button("Cancel", variant="error", id="cancel"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            self.dismiss(True)
        else:
            self.dismiss(False)



from textual.widgets import RichLog


class Log(RichLog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.display = False

    def on_mount(self):
        self.styles.border = ("round", "#0178D4")
        self.border_title = "Logs"
